import os

from enum import Enum
from typing import Optional
from pathlib import Path
from pydantic import Field, SecretStr, Secret, HttpUrl, RedisDsn, BaseModel
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
)
from pydantic_vault import VaultSettingsSource


class SecretRedis(Secret[RedisDsn]):
    def _display(self) -> str:
        return "redis://******@*****/**"


class BotPersistence(str, Enum):
    file: str = "file"
    redis: str = "redis"


class Persistence(BaseModel):
    TYPE: Optional[BotPersistence] = None
    FILE: Optional[Path] = "../pickle.bin"


class Settings(BaseSettings):
    LOG_LEVEL: str = "DEBUG"
    CONVERSATION_TIMEOUT: float = 300
    EVENTS_PER_LIST: int = 20
    PORT: int = 8000
    SCOPES: list = ["https://www.googleapis.com/auth/calendar"]
    CALENDAR_ID: str = "your_gcal_id@group.calendare.google.com"
    VAULT_SECRET_PATH: str = "secret/data/path/to/secret"
    YAML_CONFIG_PATH: Path = "../config.yaml"
    REDIS_DSN: Optional[SecretRedis] = Field(
        default=None,
        json_schema_extra={
            "vault_secret_path": os.getenv("VAULT_SECRET_PATH", VAULT_SECRET_PATH),
            "vault_secret_key": "REDIS_DSN",
        },
    )
    GOOGLE_CLIENT_CONFIG: SecretStr = Field(
        ...,
        json_schema_extra={
            "vault_secret_path": os.getenv("VAULT_SECRET_PATH", VAULT_SECRET_PATH),
            "vault_secret_key": "GOOGLE_CLIENT_CONFIG",
        },
    )
    TELEGRAM_BOT_TOKEN: SecretStr = Field(
        ...,
        json_schema_extra={
            "vault_secret_path": os.getenv("VAULT_SECRET_PATH", VAULT_SECRET_PATH),
            "vault_secret_key": "TELEGRAM_BOT_TOKEN",
        },
    )
    WEBHOOK_MODE: bool = False
    WEBHOOK_URL: HttpUrl = "https://tg-bot.psynet.su/"
    WEBHOOK_SECRET: SecretStr = Field(
        ...,
        json_schema_extra={
            "vault_secret_path": os.getenv("VAULT_SECRET_PATH", VAULT_SECRET_PATH),
            "vault_secret_key": "WEBHOOK_SECRET",
        },
    )
    PERSISTENCE: Optional[Persistence] = None
    DISABLE_NOTIFICATION: bool = True

    model_config: SettingsConfigDict = {
        "case_sensitive": True,
        "vault_url": os.getenv("VAULT_URL"),
        "vault_kubernetes_role": os.getenv("VAULT_KUBERNETES_ROLE", None),
        "vault_token": os.getenv("VAULT_TOKEN", None),
        "vault_auth_mount_point": os.getenv("VAULT_AUTH_MOUNT_POINT", None),
        "vault_namespace": os.getenv("VAULT_NAMESPACE", None),
        "env_nested_delimiter": "__",
    }

    @classmethod
    def settings_customise_sources(  # noqa
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        # This is where you can choose which settings sources to use and their priority
        if "VAULT_URL" in os.environ:
            return (
                init_settings,
                env_settings,
                dotenv_settings,
                VaultSettingsSource(settings_cls),
                file_secret_settings,
            )
        return (
            init_settings,
            env_settings,
            dotenv_settings,
            # VaultSettingsSource(settings_cls),
            file_secret_settings,
        )


settings = Settings()



import asyncio
import uvicorn.server
import src.telegram.tg_wrapper as TgHandlers
import uvicorn
import src.telegram.common as Common

from contextlib import asynccontextmanager
from telegram import Update
from telegram.ext import ApplicationBuilder, CommandHandler, MessageHandler, filters
from fastapi import FastAPI, Request, Response
from fastapi.responses import JSONResponse
from starlette_prometheus import metrics, PrometheusMiddleware
from http import HTTPStatus
from importlib.metadata import version

from src.configs.config import settings
from src.utils.logger import logger
from src.telegram.bot_persistence import bot_persistense_config


@asynccontextmanager
async def lifespan(_: FastAPI):
    if settings.WEBHOOK_MODE:
        logger.info(f"Running in webhook mode, webhook url: {settings.WEBHOOK_URL}")
        await tg_app.bot.setWebhook(
            url=str(settings.WEBHOOK_URL),
            secret_token=settings.WEBHOOK_SECRET.get_secret_value(),
        )  # replace <your-webhook-url>
        async with tg_app:
            await tg_app.start()
            yield
            await tg_app.stop()
    else:
        logger.info("Running in polling mode")
        await tg_app.bot.deleteWebhook()
        async with tg_app:
            await tg_app.updater.start_polling()
            await tg_app.start()
            yield
            await tg_app.updater.stop()
            await tg_app.stop()


app = FastAPI(
    lifespan=lifespan, version=version("psytican-bot"), title="Psytican Bot API"
)

app.add_middleware(PrometheusMiddleware)
app.add_route("/metrics/", metrics, name="metrics", include_in_schema=True)


@app.post("/")
async def process_update(request: Request):
    req = await request.json()
    logger.debug(f"{req=}")
    update = Update.de_json(req, tg_app.bot)
    await tg_app.process_update(update)
    return Response(status_code=HTTPStatus.OK)


@app.get("/healthz")
async def healthz(request: Request):
    response = {"version": version("psytican-bot"), "status": "ok"}
    return JSONResponse(
        content=response,
    )


@app.get("/metric")
async def metric(request: Request):
    return metrics(request)


@app.patch("/update_acls")
async def update_acls(request: Request):
    Common.update_acl()
    return Response(
        status_code=HTTPStatus.NO_CONTENT,
    )


tg_app = (
    ApplicationBuilder().token(settings.TELEGRAM_BOT_TOKEN.get_secret_value()).build()
)
if settings.PERSISTENCE:
    tg_app.persistence = bot_persistense_config(settings.PERSISTENCE.TYPE)

tg_app.add_handler(TgHandlers.conv_handler)
tg_app.add_handler(CommandHandler("start", TgHandlers.general_start))
tg_app.add_handler(CommandHandler("hello", TgHandlers.hello))
tg_app.add_handler(CommandHandler("help", TgHandlers.help_command))
tg_app.add_handler(CommandHandler("cancel", TgHandlers.cancel))
tg_app.add_handler(
    CommandHandler("update_acls", TgHandlers.update_acls, filters=Common.admin_acl)
)
tg_app.add_handler(
    MessageHandler(
        (filters.Regex(TgHandlers.MESSAGE_NEW_EVENT_GRREDY_PATTERNS))
        & (Common.chat_acl | Common.admin_acl),
        TgHandlers.fast_book,
    )
)


async def run_bot() -> None:
    logger.info("Starting bot...")
    logger.info(f"Notifications disabled: {settings.DISABLE_NOTIFICATION}")
    if settings.PERSISTENCE and settings.PERSISTENCE.TYPE:
        logger.info(f"Persistence: {settings.PERSISTENCE.TYPE.name}")
    else:
        logger.info("Persistence: disabled")

    from src.metrics.bot_info_metrics import tg_app_info  # noqa

    server = uvicorn.Server(
        config=uvicorn.Config(
            "src.main:app",
            port=int(settings.PORT),
            host="0.0.0.0",
            reload=True,
            log_config=None,
        )
    )

    async with tg_app:
        await server.serve()


def main():
    asyncio.run(run_bot())


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        logger.info("Application stopped")



from redis import Redis
from telegram.ext import PicklePersistence
from pydantic import RedisDsn
from src.configs.config import settings, BotPersistence
from src.telegram.redispesistence import RedisPersistence


def bot_persistense_config(persistence: BotPersistence):
    if persistence == "file":
        bot_persistence = PicklePersistence(
            filepath=settings.PERSISTENCE.FILE, update_interval=0
        )
    elif persistence == "redis":
        redis_dsn: RedisDsn = settings.REDIS_DSN.get_secret_value()
        redis_instance = Redis(
            host=redis_dsn.host,
            port=redis_dsn.port,
            db=redis_dsn.path[1:],
            password=redis_dsn.password,
        )
        bot_persistence = RedisPersistence(redis=redis_instance, update_interval=0)
    else:
        bot_persistence = None
    return bot_persistence



import logging
from src.configs.config import settings


class EndpointFilter(logging.Filter):
    def __init__(self, name: str = "", endpoint: str = "") -> None:
        super().__init__(name)
        self.endpoint: str = endpoint

    def filter(self, record: logging.LogRecord) -> bool:
        return record.getMessage().find(self.endpoint) == -1


logging.basicConfig(
    format="%(asctime)s - %(name)s - %(levelname)s - %(module)s: %(message)s",
    level=settings.LOG_LEVEL,
    # format="%(asctime)s - %(name)s - %(levelname)s - %(message)s", level=settings.LOG_LEVEL
)
# set higher logging level for httpx to avoid all GET and POST requests being logged
logging.getLogger("httpx").setLevel(logging.WARNING)
logging.getLogger("httpcore").setLevel(logging.INFO)
logging.getLogger("telegram.ext").setLevel(logging.INFO)
logging.getLogger("telegram.ext.ConversationHandler").setLevel(logging.INFO)
logging.getLogger("apscheduler.scheduler").setLevel(logging.INFO)
logging.getLogger("oauth2client").setLevel(logging.INFO)
logging.getLogger("pydantic-vault").setLevel(logging.DEBUG)
logging.getLogger("uvicorn.access").addFilter(EndpointFilter(endpoint="/healthz"))
logging.getLogger("uvicorn.access").addFilter(EndpointFilter(endpoint="/metrics"))
logging.getLogger("uvicorn.access").addFilter(EndpointFilter(endpoint="/metrics/"))

# logger = logging.getLogger(__name__)
logger = logging.getLogger(name="bookBot")



from src.telegram.tg_acl import ChatACL, AdminACL
from src.configs.user_conf import yaml_settings
from src.utils.logger import logger

chat_acl = ChatACL(yaml_settings.allowed_chats)
admin_acl = AdminACL(yaml_settings.admin_users)
global_filter = chat_acl | admin_acl


def update_acl():
    yaml_settings.__init__()
    logger.info("Config reread")
    logger.debug(f"{yaml_settings.admin_users=}, {yaml_settings.allowed_chats=}")
    chat_acl.update_acl(
        chats=yaml_settings.allowed_chats,
    )
    logger.info(f"Updated allowed chats: {','.join(str(yaml_settings.allowed_chats))}")
    admin_acl.update_acl(admins=yaml_settings.admin_users)
    logger.info(f"Updated admin users: {','.join(yaml_settings.admin_users)}")



import json
import re
import src.telegram.tg_calendar as Calendar
import src.telegram.common as Common

from telegram import (
    Update,
    ReplyKeyboardMarkup,
    InlineKeyboardButton,
    InlineKeyboardMarkup,
)
from telegram.ext import (
    CallbackQueryHandler,
    CommandHandler,
    ContextTypes,
    ConversationHandler,
    MessageHandler,
    filters,
)
from src.configs.config import settings
from src.google_api.gcalendar import GoogleCalendar
from src.models.calendar.event import CalendarEvent, CalendarDateTime
from src.models.calendar.event_meta import CalendarEventMetadata
from src.telegram.utils import format_calndar_date
from src.utils.logger import logger
from src.utils.convert import DATE_PATTERN, DateParser

# Top-level converstation handler callbacks
ANSWER, CANCEL = map(chr, range(2))
# Calendar  converstation handler
CAL_CONTROL, EDIT_START_DATE, EDIT_END_DATE = map(chr, range(2, 5))
# Event list conversation handler callbacks
GET_EVENTS, EVENT_MENU, DELETE_EVENT, EVENT_CONTROL = map(chr, range(5, 9))
# Event edit conversation handler callbacks
(EVENT_CREATE, EDIT_EVENT, EVENT_SAVE, EVENT_DESC, TYPING, EVENT_EDITOR) = map(
    chr, range(9, 15)
)
# Common callbacks
(CURRENT_EVENT, NEW_EVENT, NEW_EVENT_DICT, BACK, RESTART) = map(chr, range(15, 20))
END = ConversationHandler.END

# Regex patterns
MESSAGE_NEW_EVENT_GRREDY_PATTERNS = r"^([Nn]ew\ event|[Bb]ook|[Бб]ук)"
MESSAGE_NEW_EVENT_PATTERNS = MESSAGE_NEW_EVENT_GRREDY_PATTERNS + "$"
MESSAGE_GET_EVENT_PATTERNS = r"^([Gg]et\ events)$"
MESSAGE_PATTERNS = MESSAGE_NEW_EVENT_PATTERNS + "|" + MESSAGE_GET_EVENT_PATTERNS
MESSAGE_CANCEL_PATTERNS = r"^([Cc]ancel|[Ss]top)$"


async def hello(update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:
    calendars = GoogleCalendar().get_calendars()
    await update.message.reply_text(
        f"Hello {update.effective_user.first_name}",
        disable_notification=settings.DISABLE_NOTIFICATION,
    )
    await update.message.reply_text(json.dumps(calendars))
    await update.message.delete()


async def fast_book(update: Update, context: ContextTypes.DEFAULT_TYPE) -> int:
    usage_text: str = (
        "\n<b>Usage:\n</b>"
        "<code>book &lt;start_date&gt;[-&lt;end_date&gt;]: &lt;Event description&gt;</code>\n"
        "Everything in &lt;&gt; brackets is mandatory, in [] brackets is optional.\n "
        "if &lt;end date&gt; is not defined event end date will be equal it's start date.\n"
        "<b>Examples:</b>\n"
        '"<code>book 22.11: My awesome event</code>" will create the event at 24 of '
        "November this year\n"
        '"<code>бук 22.11-23.11: 2 Саба 2 верха</code>" will create the event 22 to 23 of '
        "November this year\n"
        '"<code>бук 22.11.2024: 2 Саба 2 верха</code>" will create the event 22 November 2024\n'
    )
    logger.debug(update.message.text)
    date_pattern = DATE_PATTERN
    regex = re.compile(
        MESSAGE_NEW_EVENT_GRREDY_PATTERNS
        + r"\s+("
        + date_pattern
        + r")(?:-("
        + date_pattern
        + r")?)?\:?\s*(.*)$"
    )
    logger.debug(f"{regex=}, {regex.groups=}, {regex.pattern=}, {regex.flags=}")
    matching = regex.match(update.message.text)
    if not matching:
        await update.message.reply_text(
            text="<b>Incorrect booking request</b>\n" + usage_text,
            parse_mode="HTML",
            disable_notification=settings.DISABLE_NOTIFICATION,
        )
        return END
    param_list = regex.findall(update.message.text)[0]
    logger.debug(f"{matching=}, {param_list=}")
    index_act = 0
    index_start_date = 1
    index_end_date = 5
    index_description = 9
    action = param_list[index_act]
    try:
        start_date = DateParser.parse_date(param_list[index_start_date])
    except ValueError as err:
        logger.error(f"Bad start date. Error: {err}")
        await update.message.reply_text(
            text=f'<b>"{param_list[index_start_date]}" is incorrect start date.</b>\n'
            + usage_text,
            parse_mode="HTML",
            disable_notification=settings.DISABLE_NOTIFICATION,
        )
        return END
    if not param_list[index_end_date]:
        end_date = start_date
    else:
        try:
            end_date = DateParser.parse_date(param_list[index_end_date])
        except ValueError as err:
            logger.error(f"Bad end date. Error: {err}")
            await update.message.reply_text(
                text=f'<b>"{param_list[index_end_date]}" is incorrect end date</b>\n'
                + usage_text,
                parse_mode="HTML",
                disable_notification=settings.DISABLE_NOTIFICATION,
            )
            return END
    if not param_list[index_description]:
        logger.error("Event description not defined")
        await update.message.reply_text(
            text="<b>You must define the event summary.</b>\n" + usage_text,
            parse_mode="HTML",
            disable_notification=settings.DISABLE_NOTIFICATION,
        )
        return END
    description = param_list[index_description]
    logger.info(
        f'Creating event... Action: "{action}". Start date: "{start_date}". '
        f'End date: "{end_date}". Description: "{description}"'
    )
    context.user_data[NEW_EVENT] = True
    context.user_data[NEW_EVENT_DICT] = {
        "summary": description,
        "start": start_date,
        "end": end_date,
    }
    logger.info(
        f'Event "{context.user_data[NEW_EVENT_DICT]["summary"]}" which will be '
        f' held from {context.user_data[NEW_EVENT_DICT]["summary"]} to '
        f'{context.user_data[NEW_EVENT_DICT]["summary"]} is being created'
    )
    logger.debug(f"{context.user_data[NEW_EVENT_DICT]=}")
    return await save_event(update=update, context=context)


async def cancel(update: Update, context: ContextTypes.DEFAULT_TYPE) -> int:
    """Cancels and ends the conversation."""
    logger.debug(f"{update.callback_query}")
    if update.callback_query:
        user = update.callback_query.from_user
        await update.callback_query.edit_message_text(
            "Bye! I hope we can talk again some day."
        )
    else:
        user = update.message.from_user
        await update.message.reply_text(
            "Bye! I hope we can talk again some day.",
            disable_notification=settings.DISABLE_NOTIFICATION,
        )
    logger.info(f"User {user.first_name} (ID={user.id}) canceled the conversation.")
    context.user_data.pop(NEW_EVENT, None)
    context.user_data.pop(NEW_EVENT_DICT, None)
    context.user_data.pop(CURRENT_EVENT, None)
    return END


async def general_start(update: Update, context: ContextTypes.DEFAULT_TYPE) -> int:
    await update.message.reply_text("You are not allowed to communicate with me")
    logger.info(
        f"User {update.effective_user.username} ({update.effective_user.id}) is not whitelisted. "
        f"Chat ID={update.effective_chat.id}"
    )
    return END


async def update_acls(update: Update, context: ContextTypes.DEFAULT_TYPE) -> int:
    Common.update_acl()
    usernames = ",".join(map(str, Common.admin_acl.usernames))
    chats = ",".join(map(str, Common.chat_acl.chat_ids))
    logger.info(
        f'ACLs updated, admins: "{usernames}", Allowed chats: "{chats}" '
        f'by user "{update.effective_user.username}" (ID={update.effective_user.id})'
    )
    await update.message.reply_text(
        f'ACLs updated, admins: "{str(usernames)}", Allowed chats: "{chats}"',
    )


async def start(update: Update, context: ContextTypes.DEFAULT_TYPE) -> str:
    """Sends a message with three inline buttons attached."""
    logger.info(
        f'User "{update.effective_user.username}" (ID={update.effective_user.id}) '
        f'started conversation in the chat with ID "{update.effective_chat.id}"'
    )
    # logger.debug(f'{allowed_chats=}')
    keyboard = [["get events", "book", "cancel"]]

    if update.callback_query or update.message.chat.is_forum:
        inline_keyboard = [
            [
                InlineKeyboardButton(text="Get events", callback_data=str(GET_EVENTS)),
                InlineKeyboardButton(
                    text="Create event", callback_data=str(EVENT_CREATE)
                ),
                InlineKeyboardButton(text="Cancel", callback_data=str(CANCEL)),
            ]
        ]
        reply_markup = InlineKeyboardMarkup(inline_keyboard)
    else:
        reply_markup = ReplyKeyboardMarkup(
            keyboard, one_time_keyboard=False, resize_keyboard=True, is_persistent=False
        )
    logger.debug(f"{context.user_data=}")
    if update.callback_query:
        await update.callback_query.answer()
        await update.callback_query.edit_message_text(
            "Hi",
            parse_mode="HTML",
            reply_markup=reply_markup,
        )
    else:
        await update.message.reply_text(
            "Hi",
            parse_mode="HTML",
            reply_markup=reply_markup,
            disable_notification=settings.DISABLE_NOTIFICATION,
        )
    context.user_data[NEW_EVENT] = True

    return ANSWER


async def help_command(update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:
    """Displays info on how to use the bot."""
    await update.message.reply_text(
        "Use /start to test this bot.",
        disable_notification=settings.DISABLE_NOTIFICATION,
    )


async def calendar_handler(update: Update, context: ContextTypes.DEFAULT_TYPE) -> str:
    logger.info("Using calendar handler")
    logger.debug(f"{context.user_data=}")
    logger.debug(f"{str(EDIT_START_DATE)=}")
    data = update.callback_query.data
    if update.callback_query:
        if data == str(EDIT_START_DATE):
            text = "Select START date: "
            date_event = "start_date"
        elif data == str(EDIT_END_DATE):
            text = "Select END date: "
            date_event = "end_date"
        else:
            text = "Select a date: "
            date_event = "date"
        logger.debug(f"{update.callback_query.data=}")
        context.user_data["date_event"] = date_event
        await update.callback_query.answer()
        await update.callback_query.edit_message_text(
            text=text, reply_markup=Calendar.create_calendar()
        )
    else:
        await update.message.reply_text(
            text="Select a date: ",
            reply_markup=Calendar.create_calendar(),
            disable_notification=settings.DISABLE_NOTIFICATION,
        )
    return CAL_CONTROL


def event_list(events: GoogleCalendar) -> list[CalendarEvent]:
    event_matrix = []
    for event in events:
        event_txt = CalendarEvent.model_validate_json(json.dumps(event))
        event_desc = event_txt.description

        row = CalendarEvent(
            id=event_txt.id,
            summary=event_txt.summary,
            start=event_txt.start,
            end=event_txt.end,
            html_link=event_txt.html_link,
        )
        if event_desc:
            pattern = re.compile(r"((\[[^\}]{3,})?\{s*[^\}\{]{3,}?:.*\}([^\{]+\])?)")
            metadata_json = pattern.search(event_desc).group()
            logger.debug(f"{metadata_json=}")
            metadata = CalendarEventMetadata.model_validate_json(metadata_json)
            row.description = metadata
        logger.info(
            f'EVENT: Topic: "{event_txt.summary}" ' f'Date: "{event_txt.start.date}"'
        )
        logger.debug(f"{event_desc=}")
        event_matrix.append(row)
    return event_matrix


async def get_events_handler(update: Update, context: ContextTypes.DEFAULT_TYPE) -> str:
    logger.info(
        f'User "{update.effective_user.username}" (ID={update.effective_user.id}) '
        f"requested event list in chat ID={update.effective_chat.id}"
    )
    events = GoogleCalendar().get_events(
        calendar_id=settings.CALENDAR_ID, max_results=settings.EVENTS_PER_LIST
    )
    event_matrix = event_list(events=events)
    context.user_data["event_list"] = event_matrix
    logger.debug(f"{events=}")
    reply_txt = (
        f"Next {settings.EVENTS_PER_LIST} events:\n"
        f"                                                                     "
        f"                                                                 ->\n"
    )
    logger.debug(f"{reply_txt=}")
    keyboard = []
    event_list_text = []
    if len(event_matrix) > 0:
        for event in event_matrix:
            event_link = event.html_link
            logger.debug(f"Keyboard button {event.summary=}, {event_link=}")
            if event.start.date == event.end.date:
                date = event.start.date
            else:
                date = f"{event.start.date}]-[{event.end.date}"
            try:
                if event.description.owner.get("username"):
                    event_owner = {
                        "name": event.description.owner.get("username"),
                        "link": f'<a href="https://t.me/'
                        f"""{event.description.owner.get('username')}">"""
                        f"{event.description.owner.get('username')}</a>",
                    }
                else:
                    event_owner_raw = (
                        f"{event.description.owner.get('first_name')} "
                        f"{event.description.owner.get('last_name')}"
                        # f"ID: {event.description.owner.get('id')}"
                    )
                    event_owner = {"name": event_owner_raw, "link": event_owner_raw}
            except AttributeError as err:
                logger.warning(
                    f"Event {event.summary} with ID={event.id} was probably "
                    f"created manually of incorrectly. Error message {err}"
                )
                event_owner = {"name": "unknown", "link": "unknown"}
            button_text = f"[{date}]: {event.summary} ({event_owner.get("name")})"
            event_list_text.append(
                f"<b>[{date}]</b>: {event.summary} ({event_owner.get("link")})"
            )
            logger.debug(f"{event.model_dump_json()=}")
            if (
                update.effective_user.username in Common.admin_acl.usernames
                or event.description
                and event.description.owner["id"] == update.effective_user.id
            ):
                callback_data = str(EVENT_MENU) + event.id
                button_text = "✏️ " + button_text
                row = [
                    InlineKeyboardButton(text=button_text, callback_data=callback_data)
                ]
                keyboard.append(row)
            else:
                logger.warning(
                    f'Event "{event.summary}" does not have metadata '
                    f"probably it was created manually or by older version of this bot"
                )
                row = [InlineKeyboardButton(text=button_text, url=event_link)]
                keyboard.append(row)
        reply_txt = reply_txt + "\n".join(event_list_text)
    else:
        logger.info("No events found in calendar")
        reply_txt = "No upcoming events"

    logger.debug(f"{reply_txt=}")
    logger.debug(f"{update.callback_query=}")
    if update.callback_query:
        keyboard.append([InlineKeyboardButton(text="<< Back", callback_data=RESTART)])
        await update.callback_query.answer()
        await update.callback_query.edit_message_text(
            text=reply_txt,
            reply_markup=InlineKeyboardMarkup(keyboard),
            parse_mode="HTML",
            disable_web_page_preview=True,
        )
    else:
        # this pattern is used for simple messages
        keyboard.append([InlineKeyboardButton(text="Exit", callback_data=CANCEL)])
        await update.message.reply_text(
            reply_txt,
            reply_markup=InlineKeyboardMarkup(keyboard),
            parse_mode="HTML",
            disable_web_page_preview=True,
            disable_notification=settings.DISABLE_NOTIFICATION,
        )

    return EVENT_MENU


async def button(update: Update, context: ContextTypes.DEFAULT_TYPE) -> int:
    """Parses the CallbackQuery and updates the message text."""

    logger.debug(f"{update=}")
    context.user_data["answer"] = update.message.text
    logger.debug(f"{update.message.text=}")
    logger.info(
        f"{update.effective_user.username} (ID={update.effective_user.id}) "
        f'requested "{update.message.text}" in chat ID={update.effective_chat.id}"'
    )
    if re.match(MESSAGE_NEW_EVENT_PATTERNS, update.message.text):
        logger.info("Flushing event context...")
        try:
            context.user_data.pop(CURRENT_EVENT)
        except KeyError as err:
            logger.info(
                f"Context of current event ({str(CURRENT_EVENT)}) is already empety. {err}"
            )
        context.user_data[NEW_EVENT] = True
        logger.info("Creating new event")
        logger.debug(f"{update.message.text=}")
        logger.debug(f"{context.user_data=}")
        return await edit_event(update, context)
    elif re.match(MESSAGE_GET_EVENT_PATTERNS, update.message.text):
        return await get_events_handler(update=update, context=context)
    else:
        await update.message.reply_text(
            f"{update.message.text=}",
            parse_mode="HTML",
            disable_notification=settings.DISABLE_NOTIFICATION,
        )
    return ANSWER


async def inline_calendar_handler(
    update: Update, context: ContextTypes.DEFAULT_TYPE
) -> str:
    query = update.callback_query
    await query.answer()

    selected, date = await Calendar.process_calendar_selection(update, context)
    if selected:
        date_event = context.user_data.get("date_event")
        logger.debug(f"{date_event=}")
        context.user_data[date_event] = date.strftime("%d/%m/%Y")
        formatted_date = format_calndar_date(context.user_data[date_event])
        if context.user_data[NEW_EVENT]:
            if not context.user_data.get(NEW_EVENT_DICT):
                context.user_data[NEW_EVENT_DICT] = {}
            if date_event == "start_date":
                context.user_data[NEW_EVENT_DICT]["start"] = formatted_date
            elif date_event == "end_date":
                context.user_data[NEW_EVENT_DICT]["end"] = formatted_date
            else:
                context.user_data["date"] = date
        else:
            if date_event == "start_date":
                context.user_data[CURRENT_EVENT].start.date = formatted_date
            elif date_event == "end_date":
                context.user_data[CURRENT_EVENT].end.date = formatted_date
            else:
                context.user_data["date"] = date
        logger.debug(f"{context.user_data.get(CURRENT_EVENT)=}")
        logger.info(
            f"{update.effective_user.username} (ID={update.effective_user.id}) "
            f'selected {date_event} date "{date.strftime("%d.%m.%Y")}"'
        )
        await query.edit_message_text(
            text=f'You selected {date_event} {date.strftime("%d/%m/%Y")}',
        )
        logger.debug("{update=}")
        await edit_event(update, context)
    return END


async def event_menu_handler(update: Update, context: ContextTypes.DEFAULT_TYPE) -> str:
    logger.debug(f"{update.callback_query.data=}")
    logger.debug(f"{context.user_data=}")
    if re.match("^" + str(EVENT_MENU), update.callback_query.data):
        event_id = update.callback_query.data.replace(str(EVENT_MENU), "")
        logger.debug(f'Match "EVENT_MENU" ({EVENT_MENU=})')
    elif re.match("^" + str(BACK), update.callback_query.data):
        logger.debug(f'Match "BACK" ({BACK=})')
        event_id = update.callback_query.data.replace(str(BACK), "")
    logger.info(
        f'User "{update.effective_user.id}" (ID={update.effective_user.id}) '
        f'opened event menu for event "{event_id}" in chat (ID={update.effective_chat.id})'
    )
    events = context.user_data.get("event_list")
    logger.debug(f"{events=}, {event_id=}")
    for evnt in events:
        if evnt.id == event_id:
            logger.debug(f"{evnt=}")
            event: CalendarEvent = evnt
            break
    logger.debug(f"{event=}")
    if event.start.date == event.end.date:
        event_dates = event.start.date
    else:
        event_dates = f"{event.start.date} - {event.end.date}"
    logger.debug(f"{event_dates=}")
    if isinstance(event.description, CalendarEventMetadata):
        if event.description.owner.get("username"):
            event_owner = {
                "text": event.description.owner.get("username"),
                "url": f'https://t.me/{event.description.owner.get("username")}',
                "callback": None,
            }
        else:
            event_owner = {
                "text": f'{event.description.owner["first_name"]} {event.description.owner["last_name"]}',
                "url": None,
                "callback": "NO_ACTION",
            }
    else:
        event_owner = {
            "text": "NO OWNER",
            "url": f"https://t.me/{update.effective_user.username}",
            "callback": None,
        }
    context.user_data[CURRENT_EVENT] = event
    logger.debug(f"INITIALIZED CURRENT EVENT CACHE {context.user_data[CURRENT_EVENT]=}")
    keyboard = [
        [
            InlineKeyboardButton(
                f"{event_dates}: {event.summary}", url=f"{event.html_link}"
            )
        ],
        [
            InlineKeyboardButton(
                f"Created by: @{event_owner['text']}",
                url=event_owner["url"],
                callback_data=event_owner["callback"],
            )
        ],
        [
            InlineKeyboardButton("Edit", callback_data=str(EDIT_EVENT)),
            InlineKeyboardButton("Delete", callback_data=str(DELETE_EVENT)),
        ],
        [
            InlineKeyboardButton("<< Back", callback_data=str(BACK)),
            InlineKeyboardButton("Cancel", callback_data=str(CANCEL)),
        ],
    ]
    event_id = update.callback_query.data.replace(str(EVENT_MENU), "")
    context.user_data[NEW_EVENT] = False
    await update.callback_query.answer()
    await update.callback_query.edit_message_text(
        text=f"Viewing event {event_id} data:",
        reply_markup=InlineKeyboardMarkup(keyboard),
    )
    context.user_data.pop("event_list")
    return EVENT_CONTROL


async def delete_event(update: Update, context: ContextTypes.DEFAULT_TYPE) -> int:
    logger.debug(f"{context.user_data=}")
    event: CalendarEvent = context.user_data.get(CURRENT_EVENT)
    result = GoogleCalendar().delete_event(event.id)
    if result == "":
        logger.info(f"Event {event.summary} successfully deleted")
        await update.callback_query.edit_message_text(
            f"Event {event.summary} successfully deleted"
        )
    else:
        logger.error(
            f"Event {event.summary} was not deleted by the following reason: {result}"
        )
        await update.callback_query.edit_message_text(
            f"Can't delete event {event.summary}"
        )
    context.user_data.pop(CURRENT_EVENT)
    return END


async def edit_event(update: Update, context: ContextTypes.DEFAULT_TYPE) -> str:
    logger.debug(f"{context.user_data=}")
    logger.debug(f"{update.callback_query=}")
    buttons = {
        "summary": "Edit summary",
        "start": "Change start date",
        "end": "Change end date",
        "back": "<< Back",
        "callback": BACK,
        "back_button": InlineKeyboardButton("<< Back", callback_data=str(BACK)),
    }
    if context.user_data.get(NEW_EVENT):
        logger.debug("Before catch exception:")
        buttons["callback"] = RESTART
        # buttons["back"] = "Exit"
        event_dict = context.user_data.get(NEW_EVENT_DICT)
        if not event_dict:
            logger.info("No event dictionary for new event is defined.")
            event_dict = {}
        logger.debug(f"{event_dict=}")
        message = (
            f'New event <b>{event_dict.get("summary", "")}</b>:\n'
            f'<b>Start date</b>: {event_dict.get("start", "undefined")}\n'
            f'<b>End date</b>: {event_dict.get("end", "undefined")}\n'
            f"{event_dict=}"
        )
    else:
        event: CalendarEvent = context.user_data.get(CURRENT_EVENT)
        context.user_data["event_list"] = [event]
        buttons["callback"] = BACK + event.id
        logger.debug(f"{context.user_data["event_list"]=}, {event=}")
        if isinstance(event.description, CalendarEventMetadata):
            event_owner = f'@{event.description.owner.get("username")}'
        else:
            event_owner = "No owner"
        message = (
            f"Editing event\n<b>Title:</b> {event.summary}\n"
            f"<b>Start date</b>: {event.start.date}\n"
            f"<b>End date</b>: {event.end.date}\n"
            f"<b>Owner</b>: {event_owner}\n"
        )

    keyboard = [
        [
            InlineKeyboardButton(
                f"{buttons.get("summary")}", callback_data=str(EVENT_DESC)
            )
        ],
        [
            InlineKeyboardButton(
                f"{buttons.get("start")}", callback_data=str(EDIT_START_DATE)
            ),
            InlineKeyboardButton(
                f"{buttons.get("end")}", callback_data=str(EDIT_END_DATE)
            ),
        ],
        [
            InlineKeyboardButton(
                buttons.get("back"), callback_data=str(buttons.get("callback"))
            ),
            InlineKeyboardButton("Cancel", callback_data=str(CANCEL)),
            InlineKeyboardButton("Save", callback_data=str(EVENT_SAVE)),
        ],
    ]

    if update.callback_query:
        await update.callback_query.answer()
        await update.callback_query.edit_message_text(
            text=message, reply_markup=InlineKeyboardMarkup(keyboard), parse_mode="HTML"
        )
    # But after we do that, we need to send a new message
    else:
        del keyboard[2][0]
        await update.message.reply_text(
            text=message,
            reply_markup=InlineKeyboardMarkup(keyboard),
            parse_mode="HTML",
            disable_notification=settings.DISABLE_NOTIFICATION,
        )

    return EVENT_EDITOR


async def save_event(update: Update, context: ContextTypes.DEFAULT_TYPE) -> int:
    logger.debug(f"{context.user_data=}")
    if context.user_data[NEW_EVENT]:
        action = "created"
        logger.info("Creating new event")
        event_dict = context.user_data.get(NEW_EVENT_DICT, {})
        if not event_dict.get("start"):
            await update.callback_query.edit_message_text("Start date is not defined")
            return END
        logger.debug(f"{event_dict=}")
        desc = CalendarEventMetadata(
            owner=update.effective_user.to_dict()
        ).model_dump_json()
        try:
            event_body = CalendarEvent(
                summary=event_dict.get("summary"),
                description=desc,
                start=CalendarDateTime(date=event_dict.get("start")),
                end=CalendarDateTime(
                    date=event_dict.get("end", event_dict.get("start"))
                ),
            )
        except ValueError as err:
            logger.error(f"{err}")
            error_message_list = []
            for error in err.errors():
                error_message_list.append(f"<b>{str(error['ctx']['error'])}</b>")
            error_message = "\n".join(error_message_list)
            await update.callback_query.answer()
            await update.callback_query.edit_message_text(
                text=error_message, parse_mode="HTML"
            )
            return END
        logger.debug(f"{event_body=}")
        result = GoogleCalendar().create_event(event_body)
    else:
        action = "updated"
        event_body: CalendarEvent = context.user_data.get(CURRENT_EVENT)
        logger.debug(f"{event_body=}")
        try:
            CalendarEvent.model_validate(event_body)
        except ValueError as err:
            logger.error(f"{err}")
            error_message_list = []
            for error in err.errors():
                error_message_list.append(f"<b>{str(error['ctx']['error'])}</b>")
            error_message = "\n".join(error_message_list)
            await update.callback_query.answer()
            await update.callback_query.edit_message_text(
                text=error_message, parse_mode="HTML"
            )
            return END
        event_id = event_body.id
        event_body.id = None
        description = event_body.description
        try:
            event_body.description = description.model_dump_json()
        except AttributeError as err:
            logger.warning(f"Event {event_id} has bad description. Error: {err}")
        logger.debug(f"{event_id=} {event_body=}")
        result = GoogleCalendar().update_event(event_id, event_body)
    logger.debug(f"{result=}")
    logger.info(
        f'Event "{result.get('''summary''')}" {action} by {update.effective_user.username} '
        f"(ID={update.effective_user.id})"
    )
    reply_message = (
        f'Event <a href="'
        f'{result["htmlLink"]}">'
        f"{event_body.summary}</a> "
        f"successfully {action} by "
        f"<b>@{update.effective_user.username}</b>\n"
    )
    if update.callback_query:
        await update.callback_query.answer()
        await update.callback_query.edit_message_text(
            text=reply_message,
            parse_mode="HTML",
            disable_web_page_preview=True,
        )
    else:
        await update.message.reply_text(
            text=reply_message,
            parse_mode="HTML",
            disable_web_page_preview=True,
            disable_notification=settings.DISABLE_NOTIFICATION,
        )
    context.user_data.pop(NEW_EVENT, None)
    context.user_data.pop(NEW_EVENT_DICT, None)
    context.user_data.pop(CURRENT_EVENT, None)
    return END


async def ask_for_input(update: Update, context: ContextTypes.DEFAULT_TYPE) -> str:
    """Prompt user to input data for selected feature."""
    context.user_data["event_description"] = update.callback_query.data
    text = "Type the title of event or booking."

    await update.callback_query.answer()
    await update.callback_query.edit_message_text(text=text)

    return TYPING


async def save_input(update: Update, context: ContextTypes.DEFAULT_TYPE) -> str:
    """Save input for feature and return to feature selection."""
    user_data = context.user_data
    logger.debug(f"{user_data=}")
    user_data["event_description"] = update.message.text
    if context.user_data.get(NEW_EVENT):
        if not context.user_data.get(NEW_EVENT_DICT):
            context.user_data[NEW_EVENT_DICT] = {}
        logger.debug(f"{update.message.text=}")
        context.user_data[NEW_EVENT_DICT].update({"summary": update.message.text})
    else:
        user_data[CURRENT_EVENT].summary = update.message.text
    return await edit_event(update, context)


fallback_handlers = [
    CommandHandler("cancel", cancel),
    CommandHandler("start", start),
    MessageHandler(filters.Regex(MESSAGE_CANCEL_PATTERNS), cancel),
    MessageHandler(filters.Regex(MESSAGE_PATTERNS), button),
]

calendar_select_handler = ConversationHandler(
    name="calendar_select",
    persistent=True if settings.PERSISTENCE else False,
    conversation_timeout=settings.CONVERSATION_TIMEOUT,
    entry_points=[
        CallbackQueryHandler(
            calendar_handler,
            pattern=("^" + str(EDIT_START_DATE) + "$|^" + str(EDIT_END_DATE) + "$"),
        ),
        CallbackQueryHandler(inline_calendar_handler, pattern="CALENDAR"),
    ],
    states={
        CAL_CONTROL: [
            CallbackQueryHandler(inline_calendar_handler, pattern="CALENDAR"),
        ],
    },
    fallbacks=[
        CommandHandler("cancel", cancel),
        MessageHandler(filters.Regex(MESSAGE_PATTERNS), button),
    ],
    map_to_parent={
        EVENT_EDITOR: EVENT_EDITOR,
        CANCEL: CANCEL,
    },
)

conv_handler = ConversationHandler(
    name="main",
    persistent=True if settings.PERSISTENCE else False,
    conversation_timeout=settings.CONVERSATION_TIMEOUT,
    entry_points=[
        CommandHandler("start", start, filters=(Common.chat_acl | Common.admin_acl)),
        MessageHandler(
            (
                filters.Regex(MESSAGE_NEW_EVENT_PATTERNS)
                | filters.Regex(MESSAGE_GET_EVENT_PATTERNS)
            )
            & (Common.chat_acl | Common.admin_acl),
            button,
        ),
    ],
    states={
        ANSWER: [
            CallbackQueryHandler(
                get_events_handler, pattern="^" + str(GET_EVENTS) + "$"
            ),
            CallbackQueryHandler(
                edit_event,
                pattern=("^" + str(EVENT_CREATE) + "$"),
            ),
        ],
        EVENT_EDITOR: [
            CallbackQueryHandler(ask_for_input, pattern="^" + str(EVENT_DESC) + "$"),
            CallbackQueryHandler(save_event, pattern="^" + str(EVENT_SAVE) + "$"),
            CallbackQueryHandler(event_menu_handler, pattern="^" + str(BACK)),
            calendar_select_handler,
        ],
        TYPING: [MessageHandler(filters.TEXT & ~filters.COMMAND, save_input)],
        CANCEL: [CallbackQueryHandler(cancel)],
        GET_EVENTS: [
            CallbackQueryHandler(
                get_events_handler, pattern="^" + str(GET_EVENTS) + "$"
            ),
        ],
        EVENT_MENU: [
            CallbackQueryHandler(event_menu_handler, pattern="^" + str(EVENT_MENU)),
        ],
        EVENT_CONTROL: [
            CallbackQueryHandler(delete_event, pattern="^" + str(DELETE_EVENT) + "$"),
            CallbackQueryHandler(
                edit_event,
                pattern=("^" + str(EDIT_EVENT) + "$"),
            ),
            CallbackQueryHandler(get_events_handler, pattern="^" + str(BACK) + "$"),
        ],
    },
    fallbacks=fallback_handlers
    + [
        CallbackQueryHandler(cancel, pattern="^" + str(CANCEL) + "$"),
        CallbackQueryHandler(start, pattern="^" + str(RESTART) + "$"),
    ],
)



from src.configs.config import settings
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    YamlConfigSettingsSource,
)
from src.utils.convert import Convert


class YamlSettings(BaseSettings):
    TEST_STUFF: str = ""
    admin_users: list[str] | None = None
    allowed_chats: list[int] | None = None

    model_config: SettingsConfigDict = {
        "yaml_file": settings.YAML_CONFIG_PATH,
        "alias_generator": Convert.snakecase_camelcase_lower,
    }

    @classmethod
    def settings_customise_sources(  # noqa
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return (YamlConfigSettingsSource(settings_cls),)


yaml_settings = YamlSettings()



from telegram import InlineKeyboardButton, InlineKeyboardMarkup, Update
from telegram.ext import ContextTypes
import datetime
import calendar
import src.telegram.utils as utils


def create_callback_data(action, year, month, day):
    """Create the callback data associated to each button"""
    return "CALENDAR;" + ";".join([action, str(year), str(month), str(day)])


def create_calendar(year=None, month=None):
    """
    Create an inline keyboard with the provided year and month
    :param int year: Year to use in the calendar, if None the current year is used.
    :param int month: Month to use in the calendar, if None the current month is used.
    :return: Returns the InlineKeyboardMarkup object with the calendar.
    """
    now = datetime.datetime.now()
    if year is None:
        year = now.year
    if month is None:
        month = now.month
    data_ignore = create_callback_data("IGNORE", year, month, 0)
    keyboard = []
    # First row - Month and Year
    row = []
    row.append(
        InlineKeyboardButton(
            calendar.month_name[month] + " " + str(year), callback_data=data_ignore
        )
    )
    keyboard.append(row)
    # Second row - Week Days
    row = []
    for day in ["Mo", "Tu", "We", "Th", "Fr", "Sa", "Su"]:
        row.append(InlineKeyboardButton(day, callback_data=data_ignore))
    keyboard.append(row)

    my_calendar = calendar.monthcalendar(year, month)
    for week in my_calendar:
        row = []
        for day in week:
            if day == 0:
                row.append(InlineKeyboardButton(" ", callback_data=data_ignore))
            else:
                row.append(
                    InlineKeyboardButton(
                        str(day),
                        callback_data=create_callback_data("DAY", year, month, day),
                    )
                )
        keyboard.append(row)
    # Last row - Buttons
    row = []
    row.append(
        InlineKeyboardButton(
            "<", callback_data=create_callback_data("PREV-MONTH", year, month, day)
        )
    )
    row.append(InlineKeyboardButton(" ", callback_data=data_ignore))
    row.append(
        InlineKeyboardButton(
            ">", callback_data=create_callback_data("NEXT-MONTH", year, month, day)
        )
    )
    keyboard.append(row)

    return InlineKeyboardMarkup(keyboard)


async def process_calendar_selection(
    update: Update, context: ContextTypes.DEFAULT_TYPE
) -> None:
    """
    Process the callback_query. This method generates a new calendar if forward or
    backward is pressed. This method should be called inside a CallbackQueryHandler.
    :param telegram.Bot bot: The bot, as provided by the CallbackQueryHandler
    :param telegram.Update update: The update, as provided by the CallbackQueryHandler
    :return: Returns a tuple (Boolean,datetime.datetime), indicating if a date is selected
                and returning the date if so.
    """
    ret_data = (False, None)
    query = update.callback_query
    (_, action, year, month, day) = utils.separate_callback_data(query.data)
    curr = datetime.datetime(int(year), int(month), 1)
    if action == "IGNORE":
        await context.bot.answer_callback_query(callback_query_id=query.id)
    elif action == "DAY":
        await context.bot.edit_message_text(
            text=query.message.text,
            chat_id=query.message.chat_id,
            message_id=query.message.message_id,
        )
        ret_data = True, datetime.datetime(int(year), int(month), int(day))
    elif action == "PREV-MONTH":
        pre = curr - datetime.timedelta(days=1)
        await context.bot.edit_message_text(
            text=query.message.text,
            chat_id=query.message.chat_id,
            message_id=query.message.message_id,
            reply_markup=create_calendar(int(pre.year), int(pre.month)),
        )
    elif action == "NEXT-MONTH":
        ne = curr + datetime.timedelta(days=31)
        await context.bot.edit_message_text(
            text=query.message.text,
            chat_id=query.message.chat_id,
            message_id=query.message.message_id,
            reply_markup=create_calendar(int(ne.year), int(ne.month)),
        )
    else:
        await context.bot.answer_callback_query(
            callback_query_id=query.id, text="Something went wrong!"
        )
        # UNKNOWN
    return ret_data



import re
from datetime import date as _date

ERA = 2000
DATE_PATTERN = r"(\d{1,2})(?:\.|\/)(\d{1,2})(?:(?:\.|\/?)((?:\d{2}){1,2}))?"


class Convert:
    @staticmethod
    def camelcase_snakecase(value: str) -> str:
        return re.sub(r"(?<!^)(?=[A-Z])", "_", value).lower()

    @staticmethod
    def camelcase_kebabcase(value: str) -> str:
        return re.sub(r"(?<!^)(?=[A-Z])", "-", value).lower()

    @staticmethod
    def snakecase_camelcase_upper(value):
        return value.title().replace("_", "")

    @staticmethod
    def snakecase_camelcase_lower(value):
        return value[0].lower() + Convert.snakecase_camelcase_upper(value)[1:]


class DateParser:
    @staticmethod
    def parse_date(date):
        regex = re.compile(DATE_PATTERN)
        if not regex.match(date):
            raise ValueError(f'Date "{date}" does not match pattern')
        parsed_date = regex.findall(date)[0]
        day, month, year = parsed_date
        if not year:
            year = _date.today().year
        if len(str(year)) == 2:
            year = int(year) + ERA
        print(f"{day=}, {month=}, {year=}")
        return _date(year=int(year), month=int(month), day=int(day))



import datetime


def separate_callback_data(data):
    """Separate the callback data"""
    return data.split(";")


def format_calndar_date(date):
    formatted_date = datetime.datetime.strptime(date, "%d/%m/%Y").date()
    return formatted_date



from typing import List
from src.utils.logger import logger
from telegram.ext.filters import Chat


class ChatACL(Chat):
    __slots__ = ()

    def __init__(self, chat_ids=None, allow_empty: bool = True):
        super().__init__(chat_ids, allow_empty=True)
        self.allow_empty = allow_empty

    def update_acl(self, chats: List[int]):
        chat_set = set()
        if not chats:
            chats = []
        for chat in chats:
            chat_set.add(chat)
        logger.debug(f"{chat_set=}")
        super().add_chat_ids(chat_set)
        logger.info(f"update_acl: Allowed chat IDs: {self.chat_ids}")
        chat_acl_remove = set()
        for chat_id in self.chat_ids:
            if chat_id not in chat_set:
                chat_acl_remove.add(chat_id)
        logger.debug(f"{chat_acl_remove=}")
        super().remove_chat_ids(chat_acl_remove)
        logger.info(f"Chat(s) {chat_acl_remove} removed from access list")
        logger.debug(f"{self.chat_ids=}")


class AdminACL(Chat):
    __slots__ = ()

    def __init__(self, usernames=None, chat_ids=None, allow_empty: bool = False):
        super().__init__(chat_ids, usernames, allow_empty=False)
        self.allow_empty = allow_empty

    def update_acl(self, admins: List[str]):
        logger.debug(f"AdminACLs: {self.chat_ids=}")
        admin_set = set()
        if not admins:
            admins = []
        for admin in admins:
            admin_set.add(admin)
        logger.debug(f"{admin_set=}")
        super().add_usernames(admin_set)
        logger.info(f"update_acl: Admins: {self.usernames}")
        admin_acl_remove = set()
        for admin_username in self.usernames:
            if admin_username not in admin_set:
                admin_acl_remove.add(admin_username)
        logger.debug(f"{admin_acl_remove=}")
        super().remove_usernames(admin_acl_remove)
        logger.info(f"Admin(s) {admin_acl_remove} removed from admin list")
        logger.debug(f"{self.usernames=}")



from prometheus_client import Info
from importlib.metadata import version


tg_app_info = Info("bot", "Ingormation about TG_BOT")
tg_app_info.info(
    {
        "major_version": version("psytican-bot"),
        "architecture": "monolith",
    }
)



import pickle
import inspect
import logging
from collections import defaultdict
from copy import deepcopy
from typing import Any, DefaultDict, Dict, Optional, Tuple, Union, cast
from redis import Redis
from redis.exceptions import ConnectionError

from telegram.ext import BasePersistence, PersistenceInput, ContextTypes
from telegram.ext._utils.types import ConversationDict, UD, CD, CDCData, BD

logger = logging.getLogger(__name__)


class RedisPersistence(BasePersistence):
    """Using Redis to make the bot persistent"""

    def __init__(
        self,
        redis: Redis,
        on_flush: bool = False,
        store_data: PersistenceInput = None,
        update_interval: float = 60,
        context_types: ContextTypes[Any, UD, CD, BD] = None,
    ):
        super().__init__(store_data=store_data, update_interval=update_interval)
        self.redis: Redis = redis
        self.on_flush: Optional[bool] = on_flush
        self.user_data: Optional[Dict[int, UD]] = None
        self.chat_data: Optional[Dict[int, CD]] = None
        self.bot_data: Optional[BD] = None
        self.callback_data: Optional[CDCData] = None
        self.conversations: Optional[
            Dict[str, Dict[Tuple[Union[int, str], ...], object]]
        ] = None
        self.context_types: ContextTypes[Any, UD, CD, BD] = cast(
            ContextTypes[Any, UD, CD, BD], context_types or ContextTypes()
        )

    async def load_redis(self) -> None:
        try:
            get_tg_bot_presistence = self.redis.get("TelegramBotPersistence")
        except ConnectionError as err:
            logger.error(
                f"Redis is unavailable, bot persistence is disabled. "
                f"ConnectionError: {err}"
            )
            raise TypeError("Failed to connect to Redis") from err
        try:
            if inspect.iscoroutinefunction(get_tg_bot_presistence):
                data_bytes = await get_tg_bot_presistence
            else:
                data_bytes = get_tg_bot_presistence
            if data_bytes:
                data = pickle.loads(data_bytes)
                self.user_data = defaultdict(dict, data["user_data"])
                self.chat_data = defaultdict(dict, data["chat_data"])
                # For backwards compatibility with files not containing bot data
                self.bot_data = data.get("bot_data", {})
                self.callback_data = data.get("callback_data", {})
                self.conversations = data["conversations"]
            else:
                self.conversations = {}
                self.user_data = {}
                self.chat_data = {}
                self.bot_data = self.context_types.bot_data()
                self.callback_data = None
        except Exception as exc:
            raise TypeError("Something went wrong unpickling from Redis") from exc

    def dump_redis(self) -> None:
        data = {
            "conversations": self.conversations,
            "callback_data": self.callback_data,
            "user_data": self.user_data,
            "chat_data": self.chat_data,
            "bot_data": self.bot_data,
        }
        data_bytes = pickle.dumps(data)
        try:
            self.redis.set("TelegramBotPersistence", data_bytes)
        except ConnectionError as err:
            logger.error(
                f"Redis is unavailable, bot persistence is disabled. "
                f"ConnectionError: {err}"
            )

    async def get_user_data(self) -> DefaultDict[int, Dict[Any, Any]]:
        """Returns the user_data from the pickle on Redis if it exists or an empty :obj:`defaultdict`."""
        if self.user_data:
            pass
        else:
            await self.load_redis()
        return deepcopy(self.user_data)  # type: ignore[arg-type]

    async def get_chat_data(self) -> DefaultDict[int, Dict[Any, Any]]:
        """Returns the chat_data from the pickle on Redis if it exists or an empty :obj:`defaultdict`."""
        if self.chat_data:
            pass
        else:
            await self.load_redis()
        return deepcopy(self.chat_data)  # type: ignore[arg-type]

    async def get_bot_data(self) -> Dict[Any, Any]:
        """Returns the bot_data from the pickle on Redis if it exists or an empty :obj:`dict`."""
        if self.bot_data:
            pass
        else:
            await self.load_redis()
        return deepcopy(self.bot_data)  # type: ignore[arg-type]

    async def get_conversations(self, name: str) -> ConversationDict:
        """Returns the conversations from the pickle on Redis if it exsists or an empty dict."""
        if self.conversations:
            pass
        else:
            await self.load_redis()
        return self.conversations.get(name, {}).copy()  # type: ignore[union-attr]

    async def update_conversation(
        self, name: str, key: Tuple[int, ...], new_state: Optional[object]
    ) -> None:
        """Will update the conversations for the given handler and depending
        on :attr:`on_flush` save the pickle on Redis."""
        if not self.conversations:
            self.conversations = dict()
        if self.conversations.setdefault(name, {}).get(key) == new_state:
            return
        self.conversations[name][key] = new_state
        if not self.on_flush:
            self.dump_redis()

    async def update_user_data(self, user_id: int, data: Dict) -> None:
        """Will update the user_data and depending on :attr:`on_flush` save the pickle on Redis."""
        if self.user_data is None:
            self.user_data = defaultdict(dict)
        if self.user_data.get(user_id) == data:
            return
        self.user_data[user_id] = data
        if not self.on_flush:
            self.dump_redis()

    async def update_chat_data(self, chat_id: int, data: Dict) -> None:
        """Will update the chat_data and depending on :attr:`on_flush` save the pickle on Redis."""
        if self.chat_data is None:
            self.chat_data = defaultdict(dict)
        if self.chat_data.get(chat_id) == data:
            return
        self.chat_data[chat_id] = data
        if not self.on_flush:
            self.dump_redis()

    async def update_bot_data(self, data: Dict) -> None:
        """Will update the bot_data and depending on :attr:`on_flush` save the pickle on Redis."""
        if self.bot_data == data:
            return
        self.bot_data = data.copy()
        if not self.on_flush:
            self.dump_redis()

    async def flush(self) -> None:
        """Will save all data in memory to pickle on Redis."""
        self.dump_redis()

    async def drop_chat_data(self, chat_id: int) -> None:
        """Will delete the specified key from the ``chat_data`` and save the pickle file.
        Args:
            chat_id (:obj:`int`): The chat id to delete from the persistence.
        """
        if self.chat_data is None:
            return
        self.chat_data[chat_id] = None

        if not self.on_flush:
            self.dump_redis()

    async def drop_user_data(self, user_id: int) -> None:
        """Will delete the specified key from the ``user_data`` and save the pickle on Redis.
        Args:
            user_id (:obj:`int`): The user id to delete from the persistence.
        """
        if self.user_data is None:
            return
        self.user_data[user_id] = None

        if not self.on_flush:
            self.dump_redis()

    async def get_callback_data(self) -> Optional[CDCData]:
        """Returns the callback data from the pickle file if it exists or :obj:`None`.

        Returns:
            Tuple[List[Tuple[:obj:`str`, :obj:`float`, Dict[:obj:`str`, :class:`object`]]],
            Dict[:obj:`str`, :obj:`str`]] | :obj:`None`: The restored metadata or :obj:`None`,
            if no data was stored.
        """
        if self.callback_data:
            pass
        else:
            await self.load_redis()
        if self.callback_data is None:
            return None
        return deepcopy(self.callback_data)

    async def refresh_bot_data(self, bot_data: BD) -> None:
        """Does nothing."""

    async def refresh_chat_data(self, chat_id: int, chat_data: CD) -> None:
        """Does nothing."""

    async def refresh_user_data(self, user_id: int, user_data: UD) -> None:
        """Does nothing."""

    async def update_callback_data(self, data: CDCData) -> None:
        """Will update the callback_data (if changed) and save the pickle on Redis.
        Args:
            data (Tuple[List[Tuple[:obj:`str`, :obj:`float`, \
                Dict[:obj:`str`, :class:`object`]]], Dict[:obj:`str`, :obj:`str`]]):
                The relevant data to restore :class:`telegram.ext.CallbackDataCache`.
        """
        if self.callback_data == data:
            return
        self.callback_data = data
        if not self.on_flush:
            self.dump_redis()



from pydantic import BaseModel, ConfigDict
from src.utils.convert import Convert


class BaseEventModel(BaseModel):
    model_config = ConfigDict(
        use_enum_values=True,
        populate_by_name=True,
        alias_generator=Convert.snakecase_camelcase_lower,
        # arbitrary_types_allowed=True
    )



from datetime import datetime, date as _date
from typing import Optional
from typing_extensions import Self
from pydantic import Field, field_validator, model_validator
from src.models.base_schema import BaseEventModel
from src.models.calendar.event_meta import CalendarEventMetadata


class CalendarDateTime(BaseEventModel):
    date: Optional[_date] = None
    date_time: Optional[datetime] = Field(
        None,
        alias="dateTime",
    )
    time_zone: Optional[str] = None

    @model_validator(mode="after")
    def disallow_both_date_and_date_time(self):
        date = self.date
        date_time = self.date_time
        if date is not None and date_time is not None:
            raise ValueError('"date" and "dateTime" can\'t be defined simultaneously')
        return self


class CalendarEvent(BaseEventModel):
    id: Optional[str] = None
    summary: Optional[str] = None
    description: Optional[str | CalendarEventMetadata] = None
    start: Optional[CalendarDateTime]
    end: Optional[CalendarDateTime]
    html_link: Optional[str] = None

    @field_validator("description")
    def empty_str_to_none(cls, description):
        if description == "":
            return None
        return description

    @model_validator(mode="after")
    def model_validator(self) -> Self:
        start_date: _date = self.start.date
        end_date: _date = self.end.date
        if end_date is not None and end_date < start_date:
            raise ValueError("End date must be greater than start date")
        if self.end.date_time is not None and self.end.date_time < self.start.date_time:
            raise ValueError("End time must be greater than start time")
        return self



import datetime
import json

from oauth2client.service_account import ServiceAccountCredentials
from googleapiclient.discovery import build
from googleapiclient.discovery_cache.base import Cache
from googleapiclient.errors import HttpError
from src.configs.config import settings
from src.models.calendar.event import CalendarEvent
from src.utils.logger import logger


class CredsMemoryCache(Cache):
    _CACHE = {}

    def get(self, url):
        return CredsMemoryCache._CACHE.get(url)

    def set(self, url, content):
        CredsMemoryCache._CACHE[url] = content


class GoogleCalendar:
    def __init__(self):
        creds = ServiceAccountCredentials.from_json_keyfile_dict(
            json.loads(settings.GOOGLE_CLIENT_CONFIG.get_secret_value(), strict=False)
        )
        try:
            self.service = build(
                "calendar", "v3", credentials=creds, cache=CredsMemoryCache()
            )
        except HttpError as error:
            logger.error(f"An error occurred: {error}")

    def get_calendars(self):
        calendar_list = self.service.calendarList().list().execute()
        calendars = calendar_list.get("items", [])
        logger.debug(f"{calendars=}")  # TODO Loggers
        return calendars

    def get_events(self, calendar_id=None, max_results=10):
        now = (
            datetime.datetime.now(datetime.UTC) - datetime.timedelta(days=1)
        ).isoformat()
        logger.info("Getting the upcoming 10 events")
        events_result = (
            self.service.events()
            .list(
                calendarId=calendar_id,
                timeMin=now,
                maxResults=max_results,
                singleEvents=True,
                orderBy="startTime",
            )
            .execute()
        )
        events = events_result.get("items", [])
        logger.debug(f"{events=}")
        return events

    def create_event(self, event: CalendarEvent):
        event_json = json.loads(event.model_dump_json(exclude_none=True))
        logger.debug(f"{event_json=}")
        event = (
            self.service.events()
            .insert(calendarId=settings.CALENDAR_ID, body=event_json)
            .execute()
        )
        return event
        # pass

    def delete_event(self, event_id: str):
        logger.info(f"Deleting event {event_id}")
        try:
            response = (
                self.service.events()
                .delete(
                    calendarId=settings.CALENDAR_ID,
                    eventId=event_id,
                )
                .execute()
            )
            logger.debug(f"{response=}")
        except HttpError as error:
            logger.error(f"Error while deleting event {event_id}: {error}")
            return None
        logger.info(f"Event {event_id} deleted")
        return response

    def update_event(self, event_id: str, body: CalendarEvent):
        logger.info(f"Updating event {event_id}")
        body_json = json.loads(body.model_dump_json(exclude_none=True, exclude="id"))
        logger.debug(f"{body_json=}")
        # try:
        response = (
            self.service.events()
            .update(calendarId=settings.CALENDAR_ID, eventId=event_id, body=body_json)
            .execute()
        )
        logger.debug(f"{response=}")
        # except HttpError as error:
        #     logger.error(f'Error while updated event {event_id}: {error}')
        #     return None
        logger.info(f"Event {event_id} updated")
        return response
